class Appfirst::Client < Cistern::Service
  model_path "appfirst/models"
  request_path "appfirst/requests"

  #collection :servers
  collection :server_tags

  model :server
  model :server_tag

  request :create_server_tag
  #request :get_server
  #request :get_server_tags
  #request :get_servers
  #request :get_servers_tags
  #request :get_tag_servers
  #request :set_servers_tags

  recognizes :url, :logger, :adapter, :connection_options, :user, :password, :path

  module Shared
    attr_reader :url, :user, :path

    def setup(options)
      @url  = options[:url] || ENV["APPFIRST_URL"] || "https://engineyard.appfirst.com"
      @path = options[:path] || "/api/v3"

      # For auth
      @user     = options[:user]
      @password = options[:password]

      @logger = options[:logger] || Logger.new(nil)
    end
  end

  class Real
    include Shared

    def initialize(options={})
      setup(options)
      adapter            = options[:adapter] || Faraday.default_adapter
      connection_options = options[:connection_options] || {ssl: {verify: false}}

      @connection = Faraday.new({url: @url}.merge(connection_options)) do |builder|
        builder.basic_auth @user, @password
        builder.request  :url_encoded
        builder.response :json, content_type: /\bjson$/
        builder.adapter  adapter
        builder.request :retry, max: 10, interval: 0.05, interval_randomness: 0.5, backoff_factor: 2

        builder.use Appfirst::Logger, @logger

        builder.adapter *adapter
      end
    end

    def request(method, request_path, params = nil)
      response = @connection.send(method) do |req|
        req.url(File.join(@url, path, request_path, "/"))
        req.params.merge!(params)
      end

      Appfirst::Response.new(response.status, response.headers, response.body).raise!
    end
  end # Real

  class Mock
    include Shared

    def self.data
      @data ||= Hash.new do |h,k|
        {
          :servers     => {},
          :server_tags => {},
        }
      end
    end

    def data
      self.class.data[self.url]
    end

    def self.reset!
      @data = nil
      @serial_id = 0
    end

    def initialize(options={})
      setup(options)
    end

    def response(options={})
      status  = options[:status] || 200
      body    = options[:body]
      headers = {
        "Content-Type"  => "application/json; charset=utf-8"
      }.merge(options[:headers] || {})

      Appfirst::Response.new(status, headers, body).raise!
    end

    def url_for(path)
      File.join(@url.to_s, path.to_s)
    end

    def stringify_keys(hash)
      hash.is_a?(Hash) ? hash.inject({}){|r,(k,v)| r.merge(k.to_s => stringify_keys(v))} : hash
    end

    def resource_uri(fragment)
      File.join(path, fragment)
    end

    def require_parameter(params, param, options={})
      if value = params[param]
        value
      else
        errors = [{"message" => (options[:message] || "Missing parameter: #{param}"), "code" => (options[:code] || -1)}]

        response(
          :body   => {"errors" => errors},
          :status => 400,
        )
      end
    end

    def serial_id
      @@serial_id ||= 0
      @@serial_id += 1
    end
  end # Mock
end # Appfirst::Client
